"""
    Arcane - A secure remote desktop application for Windows with the
    particularity of having a server entirely written in PowerShell and
    a cross-platform client (Python/QT6).

    Author: Jean-Pierre LESUEUR (@DarkCoderSc)
    License: Apache License 2.0
    https://github.com/PhrozenIO
    https://github.com/DarkCoderSc
    https://twitter.com/DarkCoderSc
    www.phrozen.io

    Todo:
        - Find a workaround for the Wayland issue for positioning the window.
        - Offer connection profile (Add/Edit/Delete)
        - Implement Clipboard sharing between client and server.
        - Offer possibility to change the default theme with something retro and cool.
"""

import logging
import sys

from PyQt6.QtGui import QColor, QIcon, QPalette
from PyQt6.QtWidgets import QApplication

import arcane_viewer.arcane as arcane
import arcane_viewer.ui.forms as arcane_forms


def main():
    logging.basicConfig()

    logging.root.setLevel(logging.NOTSET)

    logging.basicConfig(level=logging.NOTSET)

    app = QApplication(sys.argv)

    app.setWindowIcon(QIcon(arcane.APP_ICON))

    app.setStyle('Fusion')

    # Create and apply a dark palette
    dark_palette = QPalette()
    dark_palette.setColor(QPalette.ColorRole.Window, QColor(53, 53, 53))
    dark_palette.setColor(QPalette.ColorRole.WindowText, QColor(255, 255, 255))
    dark_palette.setColor(QPalette.ColorRole.Base, QColor(42, 42, 42))
    dark_palette.setColor(QPalette.ColorRole.AlternateBase, QColor(66, 66, 66))
    dark_palette.setColor(QPalette.ColorRole.ToolTipBase, QColor(255, 255, 255))
    dark_palette.setColor(QPalette.ColorRole.ToolTipText, QColor(255, 255, 255))
    dark_palette.setColor(QPalette.ColorRole.Text, QColor(255, 255, 255))
    dark_palette.setColor(QPalette.ColorRole.Button, QColor(53, 53, 53))
    dark_palette.setColor(QPalette.ColorRole.ButtonText, QColor(255, 255, 255))
    dark_palette.setColor(QPalette.ColorRole.BrightText, QColor(255, 0, 0))
    dark_palette.setColor(QPalette.ColorRole.Link, QColor(42, 130, 218))
    dark_palette.setColor(QPalette.ColorRole.Highlight, QColor(42, 130, 218))
    dark_palette.setColor(QPalette.ColorRole.HighlightedText, QColor(0, 0, 0))

    app.setPalette(dark_palette)

    # Apply custom stylesheet
    app.setStyleSheet("""
            QGroupBox::title {
                background-color: rgb(42, 42, 42);
                padding: 4px 8px 4px 8px;
                border-radius: 4px;
                left: 8px;
                top: 8px;
            }
        """)

    # Create and show the connect window
    connect_window = arcane_forms.ConnectWindow()
    connect_window.show()

    sys.exit(app.exec())


if __name__ == '__main__':
    main()



"""
    Arcane - A secure remote desktop application for Windows with the
    particularity of having a server entirely written in PowerShell and
    a cross-platform client (Python/QT6).

    Author: Jean-Pierre LESUEUR (@DarkCoderSc)
    License: Apache License 2.0
    https://github.com/PhrozenIO
    https://github.com/DarkCoderSc
    https://twitter.com/DarkCoderSc
    www.phrozen.io
"""

from PyQt6.QtCore import Qt
from PyQt6.QtWidgets import (QComboBox, QDialog, QHBoxLayout, QLabel,
                             QPushButton, QVBoxLayout)

import arcane_viewer.ui.utilities as utilities


class ScreenSelectionWindow(QDialog, utilities.CenterWindow):
    """ Screen Selection Dialog """
    def __init__(self, parent, screens):
        super().__init__(parent)

        self.screens = screens

        self.setWindowTitle(f"Screen Selection ({len(self.screens)} Available)")

        self.setWindowFlags(
            Qt.WindowType.Dialog |
            Qt.WindowType.CustomizeWindowHint |
            Qt.WindowType.MSWindowsFixedSizeDialogHint
        )

        core_layout = QVBoxLayout()
        self.setLayout(core_layout)

        # Screen Selection Combo
        screen_selection_label = QLabel("Select Screen:")
        core_layout.addWidget(screen_selection_label)

        self.screen_selection_combobox = QComboBox()

        # screen_selection_combobox.addItems([screen.name for screen in screens])
        for screen in self.screens:
            self.screen_selection_combobox.addItem(
                screen.get_display_name(),
                userData=screen.name
            )

        core_layout.addWidget(self.screen_selection_combobox)

        core_layout.addSpacing(8)

        # Action Buttons
        action_buttons_layout = QHBoxLayout()
        core_layout.addLayout(action_buttons_layout)

        self.cancel_button = QPushButton('Cancel')
        self.cancel_button.clicked.connect(lambda: self.reject())
        action_buttons_layout.addWidget(self.cancel_button)

        self.select_button = QPushButton('Select')
        self.select_button.clicked.connect(lambda: self.accept())
        action_buttons_layout.addWidget(self.select_button)

        self.select_button.setDefault(True)
        self.select_button.setFocus()

        self.setFixedSize(290, self.sizeHint().height())

    def showEvent(self, event):
        super().showEvent(event)
        self.center_on_owner(self.parent())

    def get_selected_screen(self):
        """ Get the user-choice selected screen """
        return [screen for screen in self.screens if screen.name == self.screen_selection_combobox.currentData()][0]



"""
    Arcane - A secure remote desktop application for Windows with the
    particularity of having a server entirely written in PowerShell and
    a cross-platform client (Python/QT6).

    Author: Jean-Pierre LESUEUR (@DarkCoderSc)
    License: Apache License 2.0
    https://github.com/PhrozenIO
    https://github.com/DarkCoderSc
    https://twitter.com/DarkCoderSc
    www.phrozen.io
"""

from PyQt6.QtCore import QSize


class Screen:
    """ Screen class to store screen information """
    def __init__(self, screen_information_json: dict):
        if not all(k in screen_information_json for k in (
                "Id",
                "Name",
                "Width",
                "Height",
                "X",
                "Y",
                "Primary",
        )):
            raise Exception("Invalid screen information received")

        self.id = screen_information_json["Id"]
        self.name = screen_information_json["Name"]
        self.width = screen_information_json["Width"]
        self.height = screen_information_json["Height"]
        self.x = screen_information_json["X"]
        self.y = screen_information_json["Y"]
        self.primary = screen_information_json["Primary"]

    def get_display_name(self):
        return "#{} - {} ({}x{})".format(
            self.id,
            self.name,
            self.width,
            self.height
        )

    def size(self):
        return QSize(self.width, self.height)



"""
    Arcane - A secure remote desktop application for Windows with the
    particularity of having a server entirely written in PowerShell and
    a cross-platform client (Python/QT6).

    Author: Jean-Pierre LESUEUR (@DarkCoderSc)
    License: Apache License 2.0
    https://github.com/PhrozenIO
    https://github.com/DarkCoderSc
    https://twitter.com/DarkCoderSc
    www.phrozen.io
"""

import logging
from abc import abstractmethod

from PyQt6.QtCore import QThread, pyqtSignal, pyqtSlot

import arcane_viewer.arcane as arcane

logger = logging.getLogger(__name__)


class ClientBaseThread(QThread):
    """ Base class for all client threads """
    thread_finished = pyqtSignal(bool)

    selected_screen = None
    event_loop = None

    """`Destruction is a form of creation. So the fact they burn the money is ironic. They just want to see what happens
     when they tear the world apart. They want to change things.`, Donnie Darko"""
    def __init__(self, session: arcane.Session, worker_kind: arcane.WorkerKind):
        super().__init__()

        self._running = True
        self._connected = False
        self.session = session
        self.client = None
        self.worker_kind = worker_kind

    def run(self):
        on_error = False

        self.client = self.session.claim_client()
        try:
            logging.info("Establish New Remote Desktop Connection...")

            self.client.write_line("AttachToSession")

            self.client.write_line(self.session.session_id)

            response = self.client.read_line()
            if response != "ResourceFound":
                raise arcane.ArcaneProtocolException(arcane.ArcaneProtocolError.ResourceNotFound)

            self.client.write_line(self.worker_kind.name)

            self._connected = True

            # Implement me :-)
            self.client_execute()

            on_error = False

            logger.debug(f"`{self.__class__.__name__}` Thread gracefully ended.")
        except Exception as e:
            if self._running:
                logger.error(f"Thread `{self.__class__.__name__}` encountered an error: `{e}`")
                on_error = True
        finally:
            if self.client is not None:
                self.client.close()

            self.thread_finished.emit(on_error)

    @abstractmethod
    def client_execute(self):
        pass

    @pyqtSlot()
    def stop(self):
        self._running = False

        if self.client is not None:
            self.client.close()



"""
    Arcane - A secure remote desktop application for Windows with the
    particularity of having a server entirely written in PowerShell and
    a cross-platform client (Python/QT6).

    Author: Jean-Pierre LESUEUR (@DarkCoderSc)
    License: Apache License 2.0
    https://github.com/PhrozenIO
    https://github.com/DarkCoderSc
    https://twitter.com/DarkCoderSc
    www.phrozen.io
"""

from enum import Enum, auto

PROTOCOL_VERSION = '5.0.0b1'


class WorkerKind(Enum):
    Desktop = 0x1
    Events = 0x2


class ClipboardMode(Enum):
    Disabled = 0x1
    Receive = 0x2
    Send = 0x3
    Both = 0x4


class ArcaneProtocolCommand(Enum):
    Success = 0x1
    Fail = 0x2
    RequestSession = 0x3
    AttachToSession = 0x4
    BadRequest = 0x5
    ResourceFound = 0x6
    ResourceNotFound = 0x7
    LogonUIAccessDenied = 0x8
    LogonUIWrongSession = 0x9


class OutputEvent(Enum):
    Keyboard = 0x1
    MouseClickMove = 0x2
    MouseWheel = 0x3


class InputEvent(Enum):
    KeepAlive = 0x1
    MouseCursorUpdated = 0x2
    ClipboardUpdated = 0x3
    DesktopActive = 0x4
    DesktopInactive = 0x5


class MouseState(Enum):
    Up = 0x1
    Down = 0x2
    Move = 0x3


class MouseButton(Enum):
    Left = 0x1
    Right = 0x2
    Middle = 0x3
    Void = 0x4


class MouseCursorKind(Enum):
    IDC_APPSTARTING = auto()
    IDC_ARROW = auto()
    IDC_CROSS = auto()
    IDC_HAND = auto()
    IDC_HELP = auto()
    IDC_IBEAM = auto()
    IDC_ICON = auto()
    IDC_NO = auto()
    IDC_SIZE = auto()
    IDC_SIZEALL = auto()
    IDC_SIZENESW = auto()
    IDC_SIZENS = auto()
    IDC_SIZENWSE = auto()
    IDC_SIZEWE = auto()
    IDC_UPARROW = auto()
    IDC_WAIT = auto()


class PacketSize(Enum):
    Size1024 = 1024
    Size2048 = 2048
    Size4096 = 4096
    Size8192 = 8192
    Size9216 = 9216
    Size12288 = 12288
    Size16384 = 16384

    @property
    def display_name(self):
        return f"{self.value} bytes"


class BlockSize(Enum):
    Size32 = 32
    Size64 = 64
    Size96 = 96
    Size128 = 128
    Size256 = 256
    Size512 = 512

    @property
    def display_name(self):
        return f"{self.value}x{self.value}"



"""
    Arcane - A secure remote desktop application for Windows with the
    particularity of having a server entirely written in PowerShell and
    a cross-platform client (Python/QT6).

    Author: Jean-Pierre LESUEUR (@DarkCoderSc)
    License: Apache License 2.0
    https://github.com/PhrozenIO
    https://github.com/DarkCoderSc
    https://twitter.com/DarkCoderSc
    www.phrozen.io

    Todo:
        - (0001) : Find a way to correctly handle the Meta key from Viewer to Server especially from MacOs systems.
        - (0002) : Implement a way to handle additional special characters that are not correctly interpreted by QT.
        - (0003) : Handle shortcuts (CTRL+, WIN+, ALT+ etc..).
        - (0004) : Investigate for a generic way to handle special characters for all keyboard layouts, not just AZERTY.
"""

import logging
from sys import platform

from PyQt6.QtCore import Qt, pyqtSlot
from PyQt6.QtWidgets import QGraphicsScene, QGraphicsView

import arcane_viewer.arcane as arcane

logger = logging.getLogger(__name__)


class TangentUniverse(QGraphicsView):
    """ Virtual Desktop Host (Tangent Universe)
    `In the shadowed folds of temporal threads, how does one discern the whisper of the true path from the echo of the
    diverging veil? When the cosmic mirror distorts, do you walk the ordained spiral or the fragmented loop of the
    twilight realm?`"""

    events_thread = None
    screen = None

    def __init__(self):
        super().__init__()

        self.setHorizontalScrollBarPolicy(Qt.ScrollBarPolicy.ScrollBarAlwaysOff)
        self.setVerticalScrollBarPolicy(Qt.ScrollBarPolicy.ScrollBarAlwaysOff)

        self.setMouseTracking(True)

        self.scene = QGraphicsScene()
        self.setScene(self.scene)

    def set_event_thread(self, events_thread):
        """ Set the events thread """
        self.events_thread = events_thread

        self.events_thread.update_mouse_cursor.connect(self.update_mouse_cursor)

    def set_screen(self, screen: arcane.Screen):
        """ Set the captured screen original information """
        self.screen = screen

    def fix_mouse_position(self, x, y):
        """ Fix the virtual desktop mouse position to the original screen position """
        if self.screen is None:
            return x, y

        if self.screen.width > self.width():
            x_ratio = self.screen.width / self.width()
        else:
            x_ratio = self.width() / self.screen.width

        if self.screen.height > self.height():
            y_ratio = self.screen.height / self.height()
        else:
            y_ratio = self.height() / self.screen.height

        # We must take in account both virtual desktop size and original screen X, Y position.
        return (self.screen.x + (x * x_ratio),
                self.screen.y + (y * y_ratio))

    def send_mouse_event(self, x: int, y: int, state: arcane.MouseState, button: arcane.MouseButton):
        """ Push mouse event to the events thread """
        self.events_thread.send_mouse_event(
            x,
            y,
            state,
            button
        )

    def mouse_action_handler(self, event, is_pressed):
        """ Handle mouse press and release events """

        if self.events_thread is None:
            return

        pos = event.position()
        x, y = self.fix_mouse_position(pos.x(), pos.y())

        button = event.button()

        mouse_button = {
            Qt.MouseButton.LeftButton: arcane.MouseButton.Left,
            Qt.MouseButton.RightButton: arcane.MouseButton.Right,
            Qt.MouseButton.MiddleButton: arcane.MouseButton.Middle,
        }.get(button, arcane.MouseButton.Void)

        self.send_mouse_event(x, y, arcane.MouseState.Down if is_pressed else arcane.MouseState.Up, mouse_button)

    def mouse_click(self, event):
        """ Simulate mouse click event """
        self.mouse_action_handler(event, True)
        self.mouse_action_handler(event, False)

    def mousePressEvent(self, event):
        self.mouse_action_handler(event, True)

    def mouseReleaseEvent(self, event):
        self.mouse_action_handler(event, False)

    def mouseDoubleClickEvent(self, event):
        """ Override mouseDoubleClickEvent method to simulate a remote double click event
        Do something better than this is possible? (cross-platform) """
        self.mouse_click(event)
        self.mouse_click(event)

    def mouseMoveEvent(self, event):
        """ Override mouseMoveEvent method to handle mouse move events """
        if self.events_thread is None:
            return

        pos = event.position()
        x, y = self.fix_mouse_position(pos.x(), pos.y())

        self.send_mouse_event(x, y, arcane.MouseState.Move, arcane.MouseButton.Void)

    def keyPressEvent(self, event):
        """ Override keyPressEvent method to handle key press events """
        if self.events_thread is None:
            return

        if not event.isInputEvent():
            return

        # For a maximum of compatibility, we will not yet use match case from Python > 3.10 to handle such big enum
        # F Keys
        if event.key() == Qt.Key.Key_F1:
            key_text = "{F1}"
        elif event.key() == Qt.Key.Key_F2:
            key_text = "{F2}"
        elif event.key() == Qt.Key.Key_F3:
            key_text = "{F3}"
        elif event.key() == Qt.Key.Key_F4:
            key_text = "{F4}"
        elif event.key() == Qt.Key.Key_F5:
            key_text = "{F5}"
        elif event.key() == Qt.Key.Key_F6:
            key_text = "{F6}"
        elif event.key() == Qt.Key.Key_F7:
            key_text = "{F7}"
        elif event.key() == Qt.Key.Key_F8:
            key_text = "{F8}"
        elif event.key() == Qt.Key.Key_F9:
            key_text = "{F9}"
        elif event.key() == Qt.Key.Key_F10:
            key_text = "{F10}"
        elif event.key() == Qt.Key.Key_F11:
            key_text = "{F11}"
        elif event.key() == Qt.Key.Key_F12:
            key_text = "{F12}"
        elif event.key() == Qt.Key.Key_F13:
            key_text = "{F13}"
        elif event.key() == Qt.Key.Key_F14:
            key_text = "{F14}"
        elif event.key() == Qt.Key.Key_F15:
            key_text = "{F15}"
        elif event.key() == Qt.Key.Key_F16:
            key_text = "{F16}"

        # Arrow Keys
        elif event.key() == Qt.Key.Key_Up:
            key_text = "{UP}"
        elif event.key() == Qt.Key.Key_Down:
            key_text = "{DOWN}"
        elif event.key() == Qt.Key.Key_Left:
            key_text = "{LEFT}"
        elif event.key() == Qt.Key.Key_Right:
            key_text = "{RIGHT}"

        # Make RETURN (Numpad) key to be the same as ENTER key
        elif event.key() == Qt.Key.Key_Return or event.key() == Qt.Key.Key_Enter:
            key_text = "{ENTER}"

        # Other Special Keys
        elif event.key() == Qt.Key.Key_Backspace:
            key_text = "{BACKSPACE}"
        elif event.key() == Qt.Key.Key_Tab:
            key_text = "{TAB}"
        elif event.key() == Qt.Key.Key_Control:
            key_text = "{CTRL}"
        elif event.key() == Qt.Key.Key_Alt:
            key_text = "{ALT}"
        elif event.key() == Qt.Key.Key_Shift:
            key_text = "{SHIFT}"
        elif event.key() == Qt.Key.Key_Escape:
            key_text = "{ESC}"
        elif event.key() == Qt.Key.Key_CapsLock:
            key_text = "{CAPSLOCK}"
        elif event.key() == Qt.Key.Key_Delete:
            key_text = "{DEL}"
        elif event.key() == Qt.Key.Key_Home:
            key_text = "{HOME}"
        elif event.key() == Qt.Key.Key_End:
            key_text = "{END}"
        elif event.key() == Qt.Key.Key_PageUp:
            key_text = "{PGUP}"
        elif event.key() == Qt.Key.Key_PageDown:
            key_text = "{PGDN}"
        elif event.key() == Qt.Key.Key_Insert:
            key_text = "{INS}"
        elif event.key() == Qt.Key.Key_Help:
            key_text = "{HELP}"
        elif event.key() == Qt.Key.Key_Print:
            key_text = "{PRTSC}"
        elif event.key() == Qt.Key.Key_ScrollLock:
            key_text = "{SCROLLLOCK}"

        # TODO: 0001
        elif event.key() == Qt.Key.Key_Meta and platform != "darwin":
            key_text = "^{ESC}"

        # TODO: 0004
        # elif event.key() == Qt.Key.Key_Egrave and event.modifiers() & Qt.KeyboardModifier.AltModifier:
        #    key_text = "{`}"

        else:
            # Handle others keys that might not be correctly interpreted by PowerShell
            if event.text() in ["+", "{", "}", "%", "(", ")"]:
                key_text = f"{{{event.text()}}}"
            else:
                key_text = event.text()

        self.events_thread.send_key_event(key_text)

    def wheelEvent(self, event):
        """ Override wheelEvent method to handle mouse wheel events """
        if self.events_thread is None:
            return

        delta = event.angleDelta().y()

        self.events_thread.send_mouse_wheel_event(delta)

    @pyqtSlot(Qt.CursorShape)
    def update_mouse_cursor(self, cursor):
        self.setCursor(cursor)



"""
    Arcane - A secure remote desktop application for Windows with the
    particularity of having a server entirely written in PowerShell and
    a cross-platform client (Python/QT6).

    Author: Jean-Pierre LESUEUR (@DarkCoderSc)
    License: Apache License 2.0
    https://github.com/PhrozenIO
    https://github.com/DarkCoderSc
    https://twitter.com/DarkCoderSc
    www.phrozen.io
"""

import json
import logging

import arcane_viewer.arcane as arcane

logger = logging.getLogger(__name__)


class Session:
    option_image_quality = 80
    option_packet_size = arcane.PacketSize.Size9216
    option_block_size = arcane.BlockSize.Size64

    """ Session class to handle remote session """
    def __init__(self, server_address: str, server_port: int, password: str):
        self.server_address = server_address
        self.server_port = server_port
        self.__password = password

        self.session_id = None
        self.display_name = None
        self.presentation = False

        self.request_session()

    def claim_client(self):
        return arcane.Client(self.server_address, self.server_port, self.__password)

    def request_session(self):
        client = self.claim_client()
        try:
            client.write_line("RequestSession")

            session_information = client.read_json()
            if session_information is None:
                raise arcane.ArcaneProtocolException(arcane.ArcaneProtocolError.InvalidStructureData)

            logger.debug("@Session information:")
            logger.debug(json.dumps(session_information, indent=4))

            if not all(k in session_information for k in (
                    "SessionId",
                    "Username",
                    "MachineName",
                    "ViewOnly",
                    "Version"
            )):
                raise arcane.ArcaneProtocolException(arcane.ArcaneProtocolError.InvalidStructureData)

            if session_information["Version"] != arcane.PROTOCOL_VERSION:
                logger.error(f"Incompatible server version, client version: `{arcane.PROTOCOL_VERSION}` != "
                             f"server version: `{session_information['Version']}`")

                raise arcane.ArcaneProtocolException(
                    arcane.ArcaneProtocolError.UnsupportedVersion
                )

            self.session_id = session_information["SessionId"]

            self.display_name = "{}@{}".format(
                session_information["Username"],
                session_information["MachineName"],
            )

            if session_information["ViewOnly"]:
                logger.info("Presentation mode enforced by remote server")
                self.presentation = True

            logger.info("Session established with remote peer")
        finally:
            if client is not None:
                client.close()



"""
    Arcane - A secure remote desktop application for Windows with the
    particularity of having a server entirely written in PowerShell and
    a cross-platform client (Python/QT6).

    Author: Jean-Pierre LESUEUR (@DarkCoderSc)
    License: Apache License 2.0
    https://github.com/PhrozenIO
    https://github.com/DarkCoderSc
    https://twitter.com/DarkCoderSc
    www.phrozen.io
"""

from enum import Enum, auto


class ArcaneProtocolError(Enum):
    AuthenticationFailed = auto()
    ResourceNotFound = auto()
    InvalidStructureData = auto()
    UnsupportedVersion = auto()


class ArcaneProtocolException(Exception):
    def __init__(self, reason: ArcaneProtocolError):
        self.reason = reason

        error_messages = {
            ArcaneProtocolError.AuthenticationFailed: "Authentication failed",
            ArcaneProtocolError.ResourceNotFound: "A resource was not found",
            ArcaneProtocolError.InvalidStructureData: "Invalid structure data, whether corrupted or missing properties",
        }

        super().__init__(error_messages.get(reason, "Unknown error"))



"""
    Arcane - A secure remote desktop application for Windows with the
    particularity of having a server entirely written in PowerShell and
    a cross-platform client (Python/QT6).

    Author: Jean-Pierre LESUEUR (@DarkCoderSc)
    License: Apache License 2.0
    https://github.com/PhrozenIO
    https://github.com/DarkCoderSc
    https://twitter.com/DarkCoderSc
    www.phrozen.io
"""

from typing import Union

from PyQt6.QtWidgets import QDialog, QMainWindow


class CenterWindow:
    """ Mixin to center a window on the screen or on another window """
    def center_on_owner(self: Union[QDialog, QMainWindow], owner: Union[QDialog, QMainWindow] = None):
        if owner is None:
            owner_geometry = self.screen().availableGeometry()
        else:
            owner_geometry = owner.frameGeometry()

        subform_geometry = self.frameGeometry()

        center_point = owner_geometry.center()

        subform_geometry.moveCenter(center_point)
        self.move(subform_geometry.topLeft())



"""
    Arcane - A secure remote desktop application for Windows with the
    particularity of having a server entirely written in PowerShell and
    a cross-platform client (Python/QT6).

    Author: Jean-Pierre LESUEUR (@DarkCoderSc)
    License: Apache License 2.0
    https://github.com/PhrozenIO
    https://github.com/DarkCoderSc
    https://twitter.com/DarkCoderSc
    www.phrozen.io
"""

import sys

from PyQt6.QtCore import QT_VERSION_STR, QSize, Qt
from PyQt6.QtGui import QIcon
from PyQt6.QtWidgets import (QDialog, QHBoxLayout, QLabel, QPushButton,
                             QVBoxLayout)

import arcane_viewer.arcane as arcane
import arcane_viewer.ui.utilities as utilities


class AboutWindow(QDialog, utilities.CenterWindow):
    def __init__(self, parent):
        super().__init__(parent)

        self.setWindowTitle(f"About {arcane.APP_NAME}")

        self.setWindowFlags(
            Qt.WindowType.Dialog |
            Qt.WindowType.CustomizeWindowHint |
            Qt.WindowType.WindowCloseButtonHint |
            Qt.WindowType.MSWindowsFixedSizeDialogHint
        )

        # Setup Main Layout (Core Layout)
        core_layout = QVBoxLayout()
        self.setLayout(core_layout)

        # Avatar
        avatar = QLabel(self)
        avatar.setPixmap(QIcon(arcane.APP_ICON).pixmap(QSize(96, 96)))
        avatar.setAlignment(Qt.AlignmentFlag.AlignCenter)
        core_layout.addWidget(avatar)

        # About Author / Project
        text = QLabel(f"<h1>{arcane.APP_NAME}</h1>"
                      f"<p><b>Protocol Version:</b> {arcane.PROTOCOL_VERSION}</p>"
                      f"<p><b>Python:</b> {sys.version.split()[0]} / <b>PyQt6: </b>{QT_VERSION_STR}</p>"
                      "<p>🇫🇷 Jean-Pierre LESUEUR (<a href=\"https://twitter.com/darkcodersc\">@DarkCoderSc</a>)</p>"
                      "<p><b>License:</b> Apache License 2.0</p>"
                      "<p><a href=\"https://github.com/PhrozenIO\">www.github.com/PhrozenIO</a></p>"
                      "<p>© 2024 - <a href=\"https://www.phrozen.io\">www.phrozen.io</a></p>")
        text.setWordWrap(True)
        text.setAlignment(Qt.AlignmentFlag.AlignCenter)
        core_layout.addWidget(text)

        # Give breath to the layout (before action buttons)
        core_layout.addSpacing(8)

        # Action Button(s)
        action_buttons_layout = QHBoxLayout()
        core_layout.addLayout(action_buttons_layout)

        action_buttons_layout.addStretch(1)

        close_button = QPushButton("Close")
        close_button.clicked.connect(self.close)
        action_buttons_layout.addWidget(close_button)

        action_buttons_layout.addStretch(1)

        self.adjust_size()

    def adjust_size(self):
        self.setFixedSize(380, self.sizeHint().height())

    def showEvent(self, event):
        super().showEvent(event)
        self.center_on_owner(self.parent())



"""
    Arcane - A secure remote desktop application for Windows with the
    particularity of having a server entirely written in PowerShell and
    a cross-platform client (Python/QT6).

    Author: Jean-Pierre LESUEUR (@DarkCoderSc)
    License: Apache License 2.0
    https://github.com/PhrozenIO
    https://github.com/DarkCoderSc
    https://twitter.com/DarkCoderSc
    www.phrozen.io

    Todo:
        - (0001) : Implement remote screen resolution update, if it does, we will need to ensure it push an update of
                   selected screen object and reflect the changes where necessary.
"""

import logging

from PyQt6.QtCore import QRect, QSize, Qt, pyqtSlot
from PyQt6.QtGui import QImage, QPainter, QPixmap, QTransform
from PyQt6.QtWidgets import (QApplication, QGraphicsPixmapItem, QMainWindow,
                             QMessageBox)

import arcane_viewer.arcane as arcane
import arcane_viewer.arcane.threads as arcane_threads
import arcane_viewer.ui.custom_widgets as arcane_widgets
import arcane_viewer.ui.forms as arcane_forms

logger = logging.getLogger(__name__)


class DesktopWindow(QMainWindow):
    tangent_universe = None
    scene_pixmap = None
    v_desktop = None
    desktop_thread = None
    events_thread = None
    universe_collapsed = False

    def __init__(self, parent, session):
        super().__init__(parent)

        self.session = session

        # Set Window Properties, Layout, Title, Icon and Size
        self.setWindowTitle("🖥 {} ({}) :: {} {}".format(
            arcane.APP_NAME,
            session.server_address,
            session.display_name,
            "- View Only" if session.presentation else ""
        ))
        self.resize(QSize(640, 360))
        self.setMouseTracking(True)

        self.setContentsMargins(0, 0, 0, 0)

        self.tangent_universe = arcane_widgets.TangentUniverse()
        self.setCentralWidget(self.tangent_universe)

        self.start_desktop_thread()

    def thread_finished(self, on_error):
        """ Handle the thread finished event """
        if on_error:
            QMessageBox.critical(self, "Error", "Something went wrong, check console output for more information.")

            """ If one thread (desktop or events) failed, we will close the other one, it is considered as a critical
                error """
            self.universe_collapsed = True

        self.close()

    def start_desktop_thread(self):
        """ Start the desktop thread to handle remote desktop streaming """
        self.desktop_thread = arcane_threads.VirtualDesktopThread(self.session)
        self.desktop_thread.chunk_received.connect(self.update_scene)
        self.desktop_thread.open_cellar_door.connect(self.open_cellar_door)
        self.desktop_thread.thread_finished.connect(self.thread_finished)
        self.desktop_thread.request_screen_selection.connect(self.display_screen_selection_dialog)
        self.desktop_thread.start()

    def start_events_thread(self, screen):
        """ Start the events thread to handle remote desktop events """
        self.events_thread = arcane_threads.EventsThread(self.session)
        self.events_thread.thread_finished.connect(self.thread_finished)
        self.events_thread.start()

        # Assign our events thread to the Tangent Universe
        self.tangent_universe.set_event_thread(self.events_thread)
        self.tangent_universe.set_screen(screen)

    def close_cellar_door(self):
        """ Collapse Tangent Universe to Main Branch, We were able to save the world before 28:06:42:12 """
        if self.desktop_thread is not None:
            if self.desktop_thread.isRunning():
                self.desktop_thread.stop()
                self.desktop_thread.wait()

        if self.events_thread is not None:
            if self.events_thread.isRunning():
                self.events_thread.stop()
                self.events_thread.wait()

    def showEvent(self, event):
        super().showEvent(event)

        if self.parent() is not None:
            self.parent().hide()

    def closeEvent(self, event):
        """ Overridden close method to handle the cleanup
            `I Hope That When The World Comes To An End, I Can Breathe A Sigh Of Relief Because There Will Be So Much To
             Look Forward To.`"""

        if self.universe_collapsed:
            do_close = True
        else:
            do_close = QMessageBox.question(
                self,
                "Close",
                "Are you sure you want to close the remote desktop session?"
            ) == QMessageBox.StandardButton.Yes

        if do_close:
            self.close_cellar_door()

            event.accept()

            if self.parent() is not None:
                self.parent().show()
        else:
            event.ignore()

        super().closeEvent(event)

    def open_cellar_door(self, screen):
        """ Initialize the virtual desktop (Tangent Universe) """
        self.v_desktop = QPixmap(screen.size())
        self.v_desktop.fill(Qt.GlobalColor.black)

        self.scene_pixmap = QGraphicsPixmapItem(self.v_desktop)
        self.tangent_universe.scene.addItem(self.scene_pixmap)

        # Initialize the size of virtual desktop window regarding our current monitor screen size
        local_screen = None

        for local_screen_candidate in QApplication.screens():
            if local_screen_candidate.geometry().intersects(self.geometry()):
                local_screen = local_screen_candidate

                break

        if local_screen is None:
            local_screen = QApplication.primaryScreen()

        local_screen_size = local_screen.size()

        if local_screen_size.width() <= screen.width or local_screen_size.height() <= screen.height:
            adjust_vertically = local_screen_size.width() > local_screen_size.height()

            if adjust_vertically:
                new_width = round((local_screen_size.width() * arcane.VD_WINDOW_ADJUST_RATIO) / 100)
                resized_ratio = round(new_width * 100 / screen.width)
                new_height = round((screen.height * resized_ratio) / 100)
            else:
                new_height = round((local_screen_size.height() * arcane.VD_WINDOW_ADJUST_RATIO) / 100)
                resized_ratio = round(new_height * 100 / screen.height)
                new_width = round((screen.width * resized_ratio) / 100)

            # Resize the window
            self.setGeometry(
                local_screen.geometry().left() + (local_screen_size.width() - new_width) // 2,
                local_screen.geometry().top() + (local_screen_size.height() - new_height) // 2,
                new_width,
                new_height
            )

        # We can now start our events thread
        # TODO: 0001
        self.start_events_thread(screen)

    def fit_scene(self):
        """ Fit the scene (Hacky Technique) to the view """
        if self.tangent_universe is None or self.scene_pixmap is None:
            return

        # Instead of bellow code:
        #   `self.view.fitInView(self.scene_pixmap, Qt.AspectRatioMode.IgnoreAspectRatio)`
        # We will calculate the scale factor manually to avoid the aspect ratio issue and fitting correctly the view to
        # our virtual desktop host window.
        view_rect = self.tangent_universe.frameRect()

        scale_x = view_rect.width() / self.v_desktop.width()
        scale_y = view_rect.height() / self.v_desktop.height()

        transform = QTransform()
        transform.scale(scale_x, scale_y)
        self.tangent_universe.setTransform(transform, False)

    def update_scene(self, chunk, x, y):
        """ Update the virtual desktop with the received chunk """
        if self.v_desktop is None:
            return

        if chunk is None or not isinstance(chunk, QImage):
            return

        # Update the virtual desktop with the received chunk (Tangent Universe)
        rect = QRect(x, y, chunk.width(), chunk.height())

        painter = QPainter(self.v_desktop)
        painter.drawImage(rect, chunk)
        painter.end()

        # Update the scene with the updated virtual desktop
        self.scene_pixmap.setPixmap(self.v_desktop)

        self.fit_scene()

    def resizeEvent(self, event):
        """ Overridden resizeEvent method to fit the scene to the view """
        self.fit_scene()

    def screen_selection_rejected(self):
        self.universe_collapsed = True
        self.close()

    @pyqtSlot(list)
    def display_screen_selection_dialog(self, screens):
        """ Display screen selection dialog """
        screen_selection_dialog = arcane_forms.ScreenSelectionWindow(self, screens)
        screen_selection_dialog.accepted.connect(lambda: self.desktop_thread.on_screen_selection_dialog_closed(
            screen_selection_dialog.get_selected_screen()
        ))
        screen_selection_dialog.rejected.connect(self.screen_selection_rejected)
        screen_selection_dialog.exec()



"""
    Arcane - A secure remote desktop application for Windows with the
    particularity of having a server entirely written in PowerShell and
    a cross-platform client (Python/QT6).

    Author: Jean-Pierre LESUEUR (@DarkCoderSc)
    License: Apache License 2.0
    https://github.com/PhrozenIO
    https://github.com/DarkCoderSc
    https://twitter.com/DarkCoderSc
    www.phrozen.io
"""

import json
import os.path
import socket

from PyQt6.QtCore import Qt, pyqtSlot
from PyQt6.QtWidgets import (QComboBox, QGridLayout, QGroupBox, QHBoxLayout,
                             QLabel, QLineEdit, QMainWindow, QMessageBox,
                             QPushButton, QSpinBox, QVBoxLayout, QWidget)

import arcane_viewer.arcane as arcane
import arcane_viewer.arcane.threads as arcane_threads
import arcane_viewer.ui.forms as arcane_forms
import arcane_viewer.ui.utilities as utilities


class ConnectWindow(QMainWindow, utilities.CenterWindow):
    """ Connect Window to establish a connection to the server """

    __connect_thread = None
    __connecting_form = None
    desktop_window = None
    session = None

    def __init__(self):
        super().__init__()

        self.setWindowTitle(f"{arcane.APP_NAME} :: Connect")

        self.setWindowFlags(
            Qt.WindowType.CustomizeWindowHint |
            Qt.WindowType.WindowCloseButtonHint |
            Qt.WindowType.MSWindowsFixedSizeDialogHint
        )

        # Create central widget
        self.central_widget = QWidget()
        self.setCentralWidget(self.central_widget)

        # Setup Main Layout (Core Layout)
        core_layout = QVBoxLayout()
        self.central_widget.setLayout(core_layout)

        # Server Address Form Input
        self.server_address_label = QLabel("Server Address / Port:")
        core_layout.addWidget(self.server_address_label)

        server_address_layout = QHBoxLayout()

        self.server_address_input = QLineEdit()
        self.server_address_input.setText("127.0.0.1")
        server_address_layout.addWidget(self.server_address_input)

        separator_label = QLabel(":")
        server_address_layout.addWidget(separator_label)

        self.server_port_input = QSpinBox()
        self.server_port_input.setMinimum(0)
        self.server_port_input.setMaximum(65535)
        self.server_port_input.setValue(2801)
        server_address_layout.addWidget(self.server_port_input)

        server_address_layout.setStretch(0, 5)
        server_address_layout.setStretch(1, 0)
        server_address_layout.setStretch(2, 2)

        core_layout.addLayout(server_address_layout)

        core_layout.addSpacing(4)

        # Password Form Input
        self.password_label = QLabel("Password:")
        core_layout.addWidget(self.password_label)

        self.password_input = QLineEdit()
        self.password_input.setEchoMode(QLineEdit.EchoMode.Password)

        core_layout.addWidget(self.password_input)

        core_layout.addSpacing(4)

        # Advanced Settings (Fieldset)
        advanced_settings_group = QGroupBox("Advanced Settings")
        advanced_settings_group_layout = QGridLayout()
        advanced_settings_group.setLayout(advanced_settings_group_layout)
        core_layout.addWidget(advanced_settings_group)

        advanced_settings_group_layout.setContentsMargins(8, 16, 8, 8)

        # Virtual Desktop Image Quality
        image_quality_label = QLabel("Image Quality:")

        self.image_quality_input = QSpinBox()
        self.image_quality_input.setMinimum(10)
        self.image_quality_input.setMaximum(100)
        self.image_quality_input.setValue(80)

        # Packet Size (Optimization)
        packet_size_label = QLabel("Packet Size:")
        self.packet_size_input = QComboBox()
        for value in arcane.PacketSize:
            self.packet_size_input.addItem(value.display_name, userData=value)

        self.packet_size_input.setCurrentIndex(2)  # Default to 4096 bytes

        # Block Size (Optimization)
        block_size_label = QLabel("Block Size:")
        self.block_size_input = QComboBox()
        for value in arcane.BlockSize:
            self.block_size_input.addItem(value.display_name, userData=value)

        self.block_size_input.setCurrentIndex(1)  # Default to 64x64

        # Place Inputs in our Grid Layout
        advanced_settings_group_layout.addWidget(image_quality_label, 0, 0)
        advanced_settings_group_layout.addWidget(self.image_quality_input, 0, 1)

        advanced_settings_group_layout.addWidget(packet_size_label, 1, 0)
        advanced_settings_group_layout.addWidget(self.packet_size_input, 1, 1)

        advanced_settings_group_layout.addWidget(block_size_label, 2, 0)
        advanced_settings_group_layout.addWidget(self.block_size_input, 2, 1)

        # Give Breath to Action Buttons
        core_layout.addSpacing(6)

        # Action Buttons
        self.about_button = QPushButton("About")
        self.about_button.clicked.connect(self.show_about_dialog)

        self.cancel_button = QPushButton("Cancel")
        self.cancel_button.clicked.connect(lambda: self.close())

        self.connect_button = QPushButton("Connect")
        self.connect_button.clicked.connect(self.submit_form)
        self.connect_button.setDefault(True)

        action_layout = QHBoxLayout()
        action_layout.addWidget(self.about_button)
        action_layout.addWidget(self.cancel_button)
        action_layout.addWidget(self.connect_button)

        core_layout.addLayout(action_layout)

        # Read Default Settings: Mostly for development purposes, until custom profiles are implemented
        self.read_default()

        self.adjust_size()

    def showEvent(self, event):
        super().showEvent(event)
        self.center_on_owner()

    def read_default(self):
        """ Read default settings from the default.json file """
        if not os.path.isfile(arcane.DEFAULT_JSON):
            return

        with open(arcane.DEFAULT_JSON, "r") as f:
            try:
                data = json.load(f)
            except json.JSONDecodeError:
                return

            if "use" in data and data["use"] is False:
                return

            if "server_address" in data:
                self.server_address_input.setText(data["server_address"])

            if "server_port" in data:
                self.server_port_input.setValue(data["server_port"])

            if "server_password" in data:
                self.password_input.setText(data["server_password"])

    def submit_form(self):
        """ Validate the form and submit it """
        try:
            # Check if the ip/hostname is valid
            hostname = self.server_address_input.text()
            try:
                if socket.gethostbyname(hostname) == hostname:
                    pass
            except socket.gaierror:
                self.server_address_input.setFocus()
                raise Exception("Invalid hostname or IP address.")

            # Check password input
            if len(self.password_input.text().strip()) == 0:
                self.password_input.setFocus()
                raise Exception("Password field cannot be empty.")

            # Attempt connection
            self.__connect_thread = arcane_threads.ConnectThread(
                self.server_address_input.text(),
                self.server_port_input.value(),
                self.password_input.text(),
            )

            self.__connect_thread.thread_started.connect(self.connect_thread_started)
            self.__connect_thread.session_error.connect(self.session_error)
            self.__connect_thread.thread_finished.connect(self.connect_thread_finished)
            self.__connect_thread.start()
        except Exception as e:
            QMessageBox.critical(self, "Form Error", str(e))

    def adjust_size(self):
        self.setFixedSize(350, self.sizeHint().height())

    def show_about_dialog(self):
        about_window = arcane_forms.AboutWindow(self)
        about_window.exec()

    def get_packet_size_option(self):
        return self.packet_size_input.currentData()

    def get_block_size_option(self):
        return self.block_size_input.currentData()

    def get_image_quality_option(self):
        return self.image_quality_input.value()

    @pyqtSlot(str)
    def session_error(self, error_message):
        QMessageBox.critical(self, "Error", error_message)

    @pyqtSlot()
    def connect_thread_started(self):
        self.__connecting_form = arcane_forms.ConnectingWindow(self)
        self.__connecting_form.exec()

    @pyqtSlot(object)
    def connect_thread_finished(self, session: arcane.Session = None):
        # Close the connecting form if it is still open
        if self.__connecting_form is not None and self.__connecting_form.isVisible():
            self.__connecting_form.close()

        if session is None:
            return

        self.session = session

        # Assign Advanced Options
        self.session.option_packet_size = self.get_packet_size_option()
        self.session.option_block_size = self.get_block_size_option()
        self.session.option_image_quality = self.get_image_quality_option()

        # Show the Remote Desktop Window
        self.desktop_window = arcane_forms.DesktopWindow(self, self.session)
        self.desktop_window.show()



"""
    Arcane - A secure remote desktop application for Windows with the
    particularity of having a server entirely written in PowerShell and
    a cross-platform client (Python/QT6).

    Author: Jean-Pierre LESUEUR (@DarkCoderSc)
    License: Apache License 2.0
    https://github.com/PhrozenIO
    https://github.com/DarkCoderSc
    https://twitter.com/DarkCoderSc
    www.phrozen.io
"""

from PyQt6.QtCore import QSize, Qt
from PyQt6.QtGui import QIcon
from PyQt6.QtWidgets import (QDialog, QHBoxLayout, QLabel, QProgressBar,
                             QSizePolicy, QSpacerItem, QVBoxLayout)

import arcane_viewer.arcane as arcane
import arcane_viewer.ui.utilities as utilities


class ConnectingWindow(QDialog, utilities.CenterWindow):
    def __init__(self, parent):
        super().__init__(parent)

        self.setWindowTitle("📡 Connecting...")

        window_flags = (Qt.WindowType.Dialog |
                        Qt.WindowType.FramelessWindowHint)

        self.setWindowFlags(window_flags)

        core_layout = QHBoxLayout()
        self.setLayout(core_layout)

        # Display Icon to the Left
        icon = QLabel(self)
        icon.setPixmap(QIcon(arcane.APP_ICON).pixmap(QSize(96, 96)))
        icon.setAlignment(Qt.AlignmentFlag.AlignCenter)
        core_layout.addWidget(icon)

        # Display Information to the Right
        info_layout = QVBoxLayout()
        core_layout.addLayout(info_layout)

        spacer_top = QSpacerItem(20, 40, QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Expanding)
        info_layout.addItem(spacer_top)

        label = QLabel("Connecting to the remote server, please wait...")
        info_layout.addWidget(label)

        progress_bar = QProgressBar()
        progress_bar.setRange(0, 0)
        info_layout.addWidget(progress_bar)

        spacer_bottom = QSpacerItem(20, 40, QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Expanding)
        info_layout.addItem(spacer_bottom)

    def showEvent(self, event):
        super().showEvent(event)
        self.center_on_owner(self.parent())



"""
    Arcane - A secure remote desktop application for Windows with the
    particularity of having a server entirely written in PowerShell and
    a cross-platform client (Python/QT6).

    Author: Jean-Pierre LESUEUR (@DarkCoderSc)
    License: Apache License 2.0
    https://github.com/PhrozenIO
    https://github.com/DarkCoderSc
    https://twitter.com/DarkCoderSc
    www.phrozen.io

    Todo:
        - (0001) LogonUI Support
"""

import logging
import struct
from typing import List  # To support python <= 3.8

from PyQt6.QtCore import QByteArray, QEventLoop, pyqtSignal, pyqtSlot
from PyQt6.QtGui import QImage

import arcane_viewer.arcane as arcane

from .client_base import ClientBaseThread

logger = logging.getLogger(__name__)


class VirtualDesktopThread(ClientBaseThread):
    """ Thread to handle remote desktop streaming, at quantum level """
    open_cellar_door = pyqtSignal(arcane.Screen)
    request_screen_selection = pyqtSignal(list)
    chunk_received = pyqtSignal(QImage, int, int)

    selected_screen = None
    event_loop = None

    def __init__(self, session: arcane.Session):
        super().__init__(session, arcane.WorkerKind.Desktop)

    """`Destruction is a form of creation. So the fact they burn the money is ironic. They just want to see what happens
     when they tear the world apart. They want to change things.`, Donnie Darko"""
    def client_execute(self):
        screens_obj = self.client.read_json()
        screens = [arcane.Screen(screen) for screen in screens_obj["List"]]
        logger.info(f"{len(screens)} screen(s) detected")

        if len(screens) == 1:
            self.selected_screen = screens[0]
        else:
            self.display_screen_selection_dialog(screens)

        logger.info(f"Screen: {self.selected_screen.name} "
                    f"({self.selected_screen.width}x{self.selected_screen.height})")

        self.client.write_json(
            {
                "ScreenName": self.selected_screen.name,
                "ImageCompressionQuality": self.session.option_image_quality,
                "PacketSize": self.session.option_packet_size.value,
                "BlockSize": self.session.option_block_size.value,
                "LogonUI": False,  # TODO: 0001
            }
        )

        """ Open Cellar Door
        `This famous linguist once said, of all the phrases in the English language, of all the endless combinations
        of words in all of history, that 'cellar door' is the most beautiful.`, Karen Pomeroy"""
        self.open_cellar_door.emit(
            self.selected_screen
        )

        while self._running:
            chunk_size, x, y = struct.unpack('III', self.client.stream.read(12))
            chunk_bytes = QByteArray(self.client.stream.read(chunk_size))

            chunk = QImage()
            chunk.loadFromData(chunk_bytes)

            self.chunk_received.emit(
                chunk,
                x,
                y,
            )

    def stop(self):
        super().stop()

        if self.event_loop is not None:
            self.event_loop.quit()

    def display_screen_selection_dialog(self, screens: List[arcane.Screen]):
        self.event_loop = QEventLoop()

        self.request_screen_selection.emit(screens)

        self.event_loop.exec()

    @pyqtSlot(arcane.Screen)
    def on_screen_selection_dialog_closed(self, screen: arcane.Screen):
        if self.event_loop is not None:
            self.event_loop.quit()
            self.event_loop = None

        self.selected_screen = screen



"""
    Arcane - A secure remote desktop application for Windows with the
    particularity of having a server entirely written in PowerShell and
    a cross-platform client (Python/QT6).

    Author: Jean-Pierre LESUEUR (@DarkCoderSc)
    License: Apache License 2.0
    https://github.com/PhrozenIO
    https://github.com/DarkCoderSc
    https://twitter.com/DarkCoderSc
    www.phrozen.io
"""

import os
import sys

ASSETS_IDENTIFIER = "arcane_viewer.assets"

# Asset path getter
if sys.version_info < (3, 9):
    import pkg_resources

    def get_asset_file(asset_name):
        return pkg_resources.resource_filename(ASSETS_IDENTIFIER, asset_name)

else:
    import importlib.resources as resources

    def get_asset_file(asset_name):
        with resources.files(ASSETS_IDENTIFIER) / asset_name as asset_path:
            return str(asset_path)


# Constants or hard-coded values
APP_VERSION = "1.0.0"
APP_NAME = f"Arcane {APP_VERSION} (βeta)"
VD_WINDOW_ADJUST_RATIO = 90

DEFAULT_JSON = os.path.join(get_asset_file("default.json"))
APP_ICON = os.path.join(get_asset_file("app_icon.png"))



__author__ = "Jean-Pierre LESUEUR (@DarkCoderSc)"
__maintainer__ = "Jean-Pierre LESUEUR"
__email__ = "jplesueur@phrozen.io"
__copyright__ = "Copyright 2024, Phrozen"
__license__ = "Apache License 2.0"
__version__ = "1.0.0"
__status__ = "Development"
__date__ = "2024-07-29"

from .about import AboutWindow
from .connect import ConnectWindow
from .connecting import ConnectingWindow
from .desktop import DesktopWindow
from .screen_selection import ScreenSelectionWindow

__all__ = [
    'AboutWindow',
    'ConnectWindow',
    'ConnectingWindow',
    'DesktopWindow',
    'ScreenSelectionWindow',
]



"""
    Arcane - A secure remote desktop application for Windows with the
    particularity of having a server entirely written in PowerShell and
    a cross-platform client (Python/QT6).

    Author: Jean-Pierre LESUEUR (@DarkCoderSc)
    License: Apache License 2.0
    https://github.com/PhrozenIO
    https://github.com/DarkCoderSc
    https://twitter.com/DarkCoderSc
    www.phrozen.io
"""

import logging

from PyQt6.QtCore import QThread, pyqtSignal

import arcane_viewer.arcane as arcane

logger = logging.getLogger(__name__)


class ConnectThread(QThread):
    """ Thread to handle the first connection to the server, first-authentication and session token creation """

    thread_started = pyqtSignal()
    thread_finished = pyqtSignal(object)
    session_error = pyqtSignal(str)

    def __init__(self, server_address: str, server_port: int, password: str):
        super().__init__()

        self.server_address = server_address
        self.server_port = server_port
        self.__password = password

    def run(self):
        session = None

        self.thread_started.emit()
        try:
            session = arcane.Session(
                self.server_address,
                self.server_port,
                self.__password,
            )
        except Exception as e:
            logger.error(f"An error occurred while connecting to the server: {e}")

            if (isinstance(e, arcane.ArcaneProtocolException) and
                    e.reason == arcane.ArcaneProtocolError.AuthenticationFailed):
                error_message = "Authentication failed, check your credentials"
            elif (isinstance(e, arcane.ArcaneProtocolException) and
                    e.reason == arcane.ArcaneProtocolError.UnsupportedVersion):
                error_message = ("Protocol version mismatch, be sure to connect to a compatible server"
                                 f" (v{arcane.PROTOCOL_VERSION})")
            elif isinstance(e, TimeoutError):
                error_message = "The connection to the server timed out, check the server address and port"
            else:
                error_message = "An error occurred while connecting to the server, check the console output for more " \
                                "information."

            # Notify the main thread about the error
            self.session_error.emit(error_message)
        finally:
            self.thread_finished.emit(session)



from setuptools import find_packages, setup

setup(
    name='arcane_viewer',
    version='1.0.0b1',
    packages=find_packages(),
    include_package_data=True,
    install_requires=[
        'PyQt6',
        'setuptools',
    ],
    entry_points={
        'console_scripts': [
            'arcane-viewer = arcane_viewer.main:main',
        ],
    },
    author="Jean-Pierre LESUEUR",
    author_email="jplesueur@phrozen.io",
    description="A secure remote desktop application for Windows with the particularity of having a server entirely "
                "written in PowerShell and a cross-platform client (Python/QT6)",
    license="Apache License 2.0",
    keywords="remote desktop, remote control, remote access, remote administration, remote assistance, powershell",
    url="https://github.com/PhrozenIO/Arcane",
    python_requires='>=3.6',
)



"""
    Arcane - A secure remote desktop application for Windows with the
    particularity of having a server entirely written in PowerShell and
    a cross-platform client (Python/QT6).

    Author: Jean-Pierre LESUEUR (@DarkCoderSc)
    License: Apache License 2.0
    https://github.com/PhrozenIO
    https://github.com/DarkCoderSc
    https://twitter.com/DarkCoderSc
    www.phrozen.io

    Description:
        Client class that handles secure communication with a remote
        server using TLS. Supports password authentication through a
        challenge-based mechanism.

    Todo:
        * (0001) Add support for certificate-based authentication
        * (0002) Display dialog for unknown / first-time seen server
                 certificate
        * (0003) Support for server certificate verification
        * (0004) Implement custom exception for authentication failure,
                 display a message box
"""

import binascii
import hashlib
import json
import logging
import socket
import ssl

import arcane_viewer.arcane as arcane

logger = logging.getLogger(__name__)


class Client:
    """ Client class to handle secure communication with remote server """
    def __init__(self, server_address: str, server_port: int, password: str):
        logger.info("Connecting to remote server: `{}:{}`...".format(
            server_address,
            server_port
        ))

        self.client = socket.socket(socket.AF_INET, socket.SOCK_STREAM)

        logger.debug("Creating and configure new SSL context...")
        self.context = ssl.create_default_context()
        self.context.check_hostname = False
        self.context.verify_mode = ssl.CERT_NONE

        self.conn = self.context.wrap_socket(
            self.client,
            server_hostname=server_address,
        )

        logger.debug("Establishing connection to remote server...")
        self.conn.settimeout(10)

        self.conn.connect((server_address, server_port))

        self.conn.settimeout(None)

        logger.info(f"[{self.conn.fileno()}] Connected! Authenticating with remote server...")

        self.stream = self.conn.makefile('rwb')

        self.authenticate(password)

        logger.info("Authentication successful")

    def __del__(self):
        self.close()

    def read_line(self):
        return self.stream.readline().decode('utf-8').strip()

    def write_line(self, line: str):
        self.stream.write(line.encode('utf-8') + b'\r\n')
        self.stream.flush()

    def read_json(self):
        try:
            return json.loads(self.read_line())
        except json.JSONDecodeError:
            return None

    def write_json(self, data: dict):
        self.write_line(json.dumps(data))

    def authenticate(self, password: str):
        logger.debug("Request challenge...")

        challenge = self.read_line()

        logger.debug(f"Received challenge: `{challenge}`, attempt to solve it with defined password...")

        challenge_solution = hashlib.pbkdf2_hmac(
            "sha512",
            password.encode("utf-8"),
            challenge.encode("utf-8"),
            1000
        )
        challenge_solution = binascii.hexlify(challenge_solution)\
            .decode("utf-8").upper()

        logger.debug(f"Challenge solved: `{challenge_solution}`, sending solution to server...")

        self.write_line(challenge_solution)

        response = self.read_line()
        if response != arcane.ArcaneProtocolCommand.Success.name:
            raise arcane.ArcaneProtocolException(
                arcane.ArcaneProtocolError.AuthenticationFailed
            )

    def close(self):
        logger.info(f"[{self.conn.fileno()}] Closing connection...")

        if self.client is not None:
            try:
                self.conn.shutdown(socket.SHUT_RDWR)

                self.conn.close()
            except OSError:
                self.client.close()

                pass



"""
    Arcane - A secure remote desktop application for Windows with the
    particularity of having a server entirely written in PowerShell and
    a cross-platform client (Python/QT6).

    Author: Jean-Pierre LESUEUR (@DarkCoderSc)
    License: Apache License 2.0
    https://github.com/PhrozenIO
    https://github.com/DarkCoderSc
    https://twitter.com/DarkCoderSc
    www.phrozen.io
"""

import logging
from json.decoder import JSONDecodeError

from PyQt6.QtCore import Qt, pyqtSignal, pyqtSlot

import arcane_viewer.arcane as arcane

from .client_base import ClientBaseThread

logger = logging.getLogger(__name__)


class EventsThread(ClientBaseThread):
    update_mouse_cursor = pyqtSignal(Qt.CursorShape)

    def __init__(self, session: arcane.Session):
        super().__init__(session, arcane.WorkerKind.Events)

    def client_execute(self):
        while self._running:
            try:
                event = self.client.read_json()
            except JSONDecodeError:
                continue

            if event is None or "Id" not in event:
                continue

            event_id = event["Id"]

            if event_id == arcane.InputEvent.MouseCursorUpdated.value and "Cursor" in event:
                cursor_name = event["Cursor"]

                # Default
                cursor = Qt.CursorShape.ArrowCursor

                if cursor_name in (arcane.MouseCursorKind.IDC_SIZEALL.name,
                                   cursor_name == arcane.MouseCursorKind.IDC_SIZE.name):
                    cursor = Qt.CursorShape.SizeAllCursor
                elif cursor_name == arcane.MouseCursorKind.IDC_SIZENESW.name:
                    cursor = Qt.CursorShape.SizeBDiagCursor
                elif cursor_name == arcane.MouseCursorKind.IDC_SIZENS.name:
                    cursor = Qt.CursorShape.SizeVerCursor
                elif cursor_name == arcane.MouseCursorKind.IDC_SIZENWSE.name:
                    cursor = Qt.CursorShape.SizeFDiagCursor
                elif cursor_name == arcane.MouseCursorKind.IDC_SIZEWE.name:
                    cursor = Qt.CursorShape.SizeHorCursor
                elif cursor_name == arcane.MouseCursorKind.IDC_UPARROW.name:
                    cursor = Qt.CursorShape.UpArrowCursor
                elif cursor_name == arcane.MouseCursorKind.IDC_WAIT.name:
                    cursor = Qt.CursorShape.WaitCursor
                elif cursor_name == arcane.MouseCursorKind.IDC_APPSTARTING.name:
                    cursor = Qt.CursorShape.BusyCursor
                elif cursor_name == arcane.MouseCursorKind.IDC_CROSS.name:
                    cursor = Qt.CursorShape.CrossCursor
                elif cursor_name == arcane.MouseCursorKind.IDC_HAND.name:
                    cursor = Qt.CursorShape.PointingHandCursor
                elif cursor_name == arcane.MouseCursorKind.IDC_HELP.name:
                    cursor = Qt.CursorShape.WhatsThisCursor
                elif cursor_name == arcane.MouseCursorKind.IDC_IBEAM.name:
                    cursor = Qt.CursorShape.IBeamCursor
                elif cursor_name == arcane.MouseCursorKind.IDC_ICON.name:
                    pass  # Obsolete
                elif cursor_name == arcane.MouseCursorKind.IDC_NO.name:
                    cursor = Qt.CursorShape.ForbiddenCursor

                self.update_mouse_cursor.emit(cursor)

    @pyqtSlot(int, int, arcane.MouseState, arcane.MouseButton)
    def send_mouse_event(self, x: int, y: int, state: arcane.MouseState, button: arcane.MouseButton):
        if self.client is not None and self._connected:
            self.client.write_json(
                {
                    "Id": arcane.OutputEvent.MouseClickMove.name,
                    "X": x,
                    "Y": y,
                    "Button": button.name,
                    "Type": state.name,
                }
            )

    @pyqtSlot(str)
    def send_key_event(self, keys: str):
        if self.client is not None and self._connected:
            self.client.write_json(
                {
                    "Id": arcane.OutputEvent.Keyboard.name,
                    "Keys": keys,
                }
            )

    @pyqtSlot(int)
    def send_mouse_wheel_event(self, delta: int):
        if self.client is not None and self._connected:
            self.client.write_json(
                {
                    "Id": arcane.OutputEvent.MouseWheel.name,
                    "Delta": delta,
                }
            )



__author__ = "Jean-Pierre LESUEUR (@DarkCoderSc)"
__maintainer__ = "Jean-Pierre LESUEUR"
__email__ = "jplesueur@phrozen.io"
__copyright__ = "Copyright 2024, Phrozen"
__license__ = "Apache License 2.0"
__version__ = "1.0.0"
__status__ = "Development"
__date__ = "2024-07-29"

from .client import Client
from .constants import (APP_ICON, APP_NAME, APP_VERSION, DEFAULT_JSON,
                        VD_WINDOW_ADJUST_RATIO)
from .exceptions import ArcaneProtocolError, ArcaneProtocolException
from .protocol import (PROTOCOL_VERSION, ArcaneProtocolCommand, BlockSize,
                       ClipboardMode, InputEvent, MouseButton, MouseCursorKind,
                       MouseState, OutputEvent, PacketSize, WorkerKind)
from .screen import Screen
from .session import Session

__all__ = [
    'ArcaneProtocolError',
    'ArcaneProtocolException',
    'PROTOCOL_VERSION',
    'BlockSize',
    'ClipboardMode',
    'InputEvent',
    'MouseButton',
    'MouseCursorKind',
    'MouseState',
    'OutputEvent',
    'PacketSize',
    'ArcaneProtocolCommand',
    'WorkerKind',
    'Client',
    'Screen',
    'Session',
    'APP_ICON',
    'APP_NAME',
    'VD_WINDOW_ADJUST_RATIO',
    'APP_VERSION',
    'DEFAULT_JSON',
]
